__title__       = "pvpn"
__version__     = "0.2.2"
__license__     = "MIT"
__description__ = "Python VPN server."
__keywords__    = "vpn ike esp ip tcp udp wireguard"
__author__      = "Qian Wenjie"
__email__       = "qianwenjie@gmail.com"
__url__         = "https://github.com/qwj/python-vpn"

__all__ = ['__title__', '__version__', '__description__', '__url__']
